package com.weface.component;

import cn.hutool.core.convert.Convert;
import cn.hutool.core.lang.TypeReference;
import cn.hutool.core.util.ArrayUtil;
import cn.hutool.core.util.ObjectUtil;
import com.alibaba.fastjson.JSONArray;
import com.alibaba.fastjson.JSONObject;

import com.baomidou.mybatisplus.core.conditions.query.QueryWrapper;
import com.weface.entity.MenuTagsEntity;
import com.weface.entity.TUserTagEntity;
import com.weface.entity.UserMenusEntity;
import com.weface.service.MenuTagsService;
import com.weface.service.TUserTagService;
import com.weface.common.exception.RRException;
import com.weface.common.utils.CacheUtil;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang.StringUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;
import org.springframework.transaction.annotation.Transactional;

import java.util.*;
import java.util.concurrent.locks.ReentrantLock;
import java.util.stream.Collectors;

/**
 * @CreateUser: Administrator
 * @CreateTime: 2021/11/2
 */
@Slf4j
@Component
@Transactional
public class MenuService {

    @Autowired
    private MenuTagsService menuTagsService;
    @Autowired
    private TUserTagService tUserTagService;

    private static final CacheUtil CACHE = CacheUtil.getInstance();

    private static final ReentrantLock lock = new ReentrantLock();

    /**
     * 获取token并缓存一份
     *
     * @return 从缓存中拿token
     */
    public String getAuthToken() {
        String authToken = "";
        lock.lock();
        try {
            authToken = Convert.toStr(CACHE.getCacheData("authToken"));
            if (Objects.isNull(authToken)) {
                authToken = GeTuiUtils.getAuth();
//                authToken = "01718df19708866512701e7637fee9568a0dac5f9390609f08ef148cab56c060";
                if (ObjectUtil.isNull(authToken)) throw new RRException("获取个推token失败");
                CACHE.addCacheData("authToken", authToken);
            }
        } catch (RRException e) {
            e.printStackTrace();
            log.error("加锁失败!");
        } finally {
            lock.unlock();
        }
        return authToken;
    }

    /**
     * 通过gid以及操作系统字段获取标签code并转为实际值，后存库并返回
     *
     * @param userList 用户的giUid
     * @return
     */
    public Map<String, Object> getUserTags(String[] userList) {
        try {
            if (ArrayUtil.isEmpty(userList)) return null;
            //根据用户giuId请求到用户最新的爱好数据
            Map<String, Object> map = queryTag(userList);
            //以上均做过非空处理如果返回不为空 那么map内一定是有数据的
            if (map == null) return null;
            return map;
        } catch (Exception e) {
            e.printStackTrace();
            log.error("执行新增用户标签数据到用户表中发生异常" + e.getMessage());
        }
        return null;
    }

    /**
     * 从map中解析标签并收集关系表数据
     *
     * @param map 个推返回数据
     * @return 标签数据
     */
    public List<UserMenusEntity> getTagUser(Map<String, Object> map) {
        if (ObjectUtil.isNull(map)) return null;
        //获取所有用户标签
        List<Map<String, Object>> userTag = Convert.convert(new TypeReference<List<Map<String, Object>>>() {
        }, map.get("userTag"));
        //收集中间表数据
        List<UserMenusEntity> userMenusList = new ArrayList<>();
        //获取所有用户信息
        List<TUserTagEntity> tUserTagEntities = tUserTagService.list(new QueryWrapper<TUserTagEntity>()
                .lambda().orderByDesc(TUserTagEntity::getSaveDate));
        //遍历所有返回用户数据
        for (Map<String, Object> item : userTag) {
            //拿到当前用户giUid
            String userId = (String) item.get("userId");
            //拿到标签
            List<MenuTagsEntity> list = Convert.convert(new TypeReference<List<MenuTagsEntity>>() {
            }, item.get("tags"));
            //查找当前用户的id
            Optional<TUserTagEntity> first = tUserTagEntities.stream().filter(x -> userId.equals(x.getGid())).findFirst();
            //如果对象不为空
            if (first.isPresent()) {
                //拿到id
                String uid = first.get().getUid();
                //收集标签id
                List<Long> collect = list.stream().map(MenuTagsEntity::getId).collect(Collectors.toList());
                //遍历标签id
                Date date = new Date();
                for (Long aLong : collect) {
                    //填充数据
                    UserMenusEntity userMenusEntity = new UserMenusEntity();
                    userMenusEntity.setUserId(uid);
                    userMenusEntity.setTagsId(aLong);
                    userMenusEntity.setIsValid(1);
                    userMenusEntity.setCreateTime(date);
                    //添加数据到集合内
                    userMenusList.add(userMenusEntity);
                }
            }
        }
        return userMenusList;
    }

    /**
     * 通过用户id获取获取用户标签信息
     *
     * @param userIdList gid
     * @return
     */
    private Map<String, Object> queryTag(String[] userIdList) {
        //设置标识
        boolean flag = true;
        //获取token
        String authToken = getAuthToken();
        try {
            while (flag) {
                //将token以及用户ID封装调用画像查询接口
                String result = GeTuiUtils.queryTag(userIdList, authToken);
                JSONObject jsonObject = JSONObject.parseObject(result);
                //如果状态码为302token已过期则需要重新获取token并更新数据库token
                if (jsonObject.containsKey("error_info") && jsonObject.getJSONObject("error_info").getInteger("error_code") == 302) {
                    log.warn("token过期");
                    //authToken = "01718df19708866512701e7637fee9568a0dac5f9390609f08ef148cab56c060";
                    authToken = GeTuiUtils.getAuth();
                    if (ObjectUtil.isNull(authToken)) throw new RRException("获取个推token失败");
                    CACHE.addCacheData("authToken", authToken);
                    break;
                }
                //判断返回json内是否包含要解析的userTag
                if (jsonObject.containsKey("userTag")) {
                    JSONArray userTag = jsonObject.getJSONArray("userTag");
                    List<Map<String, Object>> userTags = getUserTagByTag(userTag);
                    Map<String, Object> all = null;
                    if (userTags != null && userTag.size() > 0) {
                        all = new HashMap<>();
                        //将数据封装并转为json格式返回
                        all.put("userTag", userTags);
                        all.put("error_code", 0);
                    }
                    return all;
                    //如果以上步骤无法走通则直接返回。并修改标识符终止循环
                } else {
                    log.error("返回值内不包含需要解析的数据" + jsonObject.toJSONString());
                    flag = false;
                }
            }
        } catch (Exception e) {
            e.printStackTrace();
            log.error("执行调用个推接口获取标签数据发生异常" + e.getMessage());
        }
        return null;
    }

    /**
     * 解析json拿到所有用户
     *
     * @param userTag 用户数据
     * @return
     */
    private List<Map<String, Object>> getUserTagByTag(JSONArray userTag) {
        List<Map<String, Object>> userTags = null;
        Integer failCount = Convert.toInt(CACHE.getCacheData("failCount"));
        Integer successCount = Convert.toInt(CACHE.getCacheData("successCount"));
        if (ObjectUtil.isNull(failCount))
            failCount = 0;
        if (ObjectUtil.isNull(successCount))
            successCount = 0;
        if (userTag != null && userTag.size() > 0) {
            //封装返回的数据userTags
            userTags = new ArrayList<>(userTag.size());
            for (Object o : userTag) {
                JSONObject jsonObject1 = JSONObject.parseObject(o.toString());
                //判断返回json内是否包含要解析的error_code，并判断返回的json状态码是否正常
                if (jsonObject1.containsKey("error_code") && jsonObject1.getInteger("error_code") == 0) {
                    String userId = "";
                    if (jsonObject1.containsKey("userId"))
                        userId = jsonObject1.getString("userId");
                    //如果从json中获取的当前giUid为空，则跳出本次循环
                    if (StringUtils.isBlank(userId)) continue;
                    //封装当前用户的所有tags信息
                    Map<String, Object> data = new HashMap<>(2);
                    //封装要返回的所有tags
                    List<MenuTagsEntity> list = getMenuByTag(jsonObject1.getJSONArray("tags"));
                    //根据giUid查询的标签如果为空则跳出本次循环
                    if (list != null && list.size() > 0) {
                        data.put("userId", userId);
                        data.put("tags", list);
                        userTags.add(data);
                    }
                    successCount++;
                } else {
                    failCount++;
                }
            }
        }
        CACHE.addCacheData("failCount", failCount);
        CACHE.addCacheData("successCount", successCount);
        return userTags;
    }

    /**
     * 解析json拿到当前用户对应的标签列表
     * 根据标签的code查询标签数据
     *
     * @param tags 标签
     * @return
     */
    private List<MenuTagsEntity> getMenuByTag(JSONArray tags) {
        List<MenuTagsEntity> list = null;
        if (tags != null && tags.size() > 0) {
            list = new ArrayList<>(tags.size());
            for (Object tag : tags) {
                JSONObject jsonObject2 = JSONObject.parseObject(tag.toString());
                String code = "";
                if (jsonObject2.containsKey("code"))
                    code = jsonObject2.getString("code");
                //如果传入标签的code为空则跳出本次循环
                if (StringUtils.isBlank(code)) continue;
                String weight = "";
                if (jsonObject2.containsKey("weight"))
                    weight = jsonObject2.getString("weight");
                //根据code或权重查询图像
                MenuTagsEntity one = menuTagsService.getOneByCodeOrWeight(code, weight);
                //如果查询的结果为空则跳出本次循环
                if (one == null) continue;
                list.add(one);
            }
        }
        return list;
    }

}
